import contextlib
import logging
import re
import os

from typing import Callable, List, Optional

import requests

from pypsrp.client import Client
from pypsrp.exceptions import WinRMError
from pypsrp.powershell import PSDataStreams

from ppa import appliance

from . import exceptions

logger = logging.getLogger(__name__)
logging.getLogger("urllib3").setLevel(logging.ERROR)
logging.getLogger("pypsrp").setLevel(logging.ERROR)

READ_TIMEOUT = 30
OPERATION_TIMEOUT = 20
KERBEROS_CACHE_PATH = "/tmp/krb5cc_0"


def _connection_wrapper(host, connection_kwargs):
    @contextlib.contextmanager
    def _open():
        try:
            psrp_client = Client(host, **connection_kwargs)
            # Getting the server config tests the connection for us.
            psrp_client.wsman.get_server_config()
            yield psrp_client

        except requests.exceptions.SSLError as e:
            logger.info(e)
            raise exceptions.CertificateError(
                "Unable to verify SSL certificate. If the certificate is "
                "self-signed, pass validate_cert=False to the initialize function"
            )

        except requests.exceptions.ConnectionError as e:
            match = re.compile(r".*(\[Errno \d+\]\s[\w\s]+)").search(str(e))
            raise exceptions.NetworkError(
                f"Connection error sending {e.request.method} "
                f"request to {e.request.url}. Error: {match.group(1) if match else None}."
            )

        finally:
            # No need to close the psrp_client.
            pass

    return _open


@appliance.min_ppa_version("2.7.0")
def _get_kerberos_session(
    *, host: str, validate_cert: bool, use_ssl: bool, read_timeout: int, operation_timeout: int,
) -> Callable:

    if not os.path.isfile(KERBEROS_CACHE_PATH):
        raise exceptions.MissingKerberosConfig(f"No Kerberos cache found at {KERBEROS_CACHE_PATH}")

    logger.info(
        f"opening psrp client to {'https' if use_ssl else 'http'}://{host}:"
        f"{'5986' if use_ssl else '5985'} using kerberos authentication"
    )
    return _connection_wrapper(
        host,
        dict(
            auth="kerberos",
            negotiate_delegate=True,
            cert_validation=validate_cert,
            ssl=use_ssl,
            read_timeout=read_timeout,
            operation_timeout=operation_timeout,
        ),
    )


def _get_ntlm_session(
    *,
    host: str,
    username: str,
    password: str,
    validate_cert: bool,
    use_ssl: bool,
    read_timeout: int,
    operation_timeout: int,
) -> Callable:
    logger.info(
        f"opening psrp client to {'https' if use_ssl else 'http'}://{host}:"
        f"{'5986' if use_ssl else '5985'} with username {username} for NTLM authentication"
    )
    return _connection_wrapper(
        host,
        dict(
            username=username,
            password=password,
            encryption="auto",
            cert_validation=validate_cert,
            ssl=use_ssl,
            read_timeout=read_timeout,
            operation_timeout=operation_timeout,
        ),
    )


def _run_dos(connection: Client, *, command: str, environment: Optional[dict] = None) -> str:
    logger.info("executing DOS command")
    std_out, std_err, _ = connection.execute_cmd(command, environment=environment)
    if std_err:
        raise exceptions.DOSCommandError(std_err)
    return std_out


def _collect_errors(streams: PSDataStreams) -> List[str]:
    """ If a powershell script makes a connection to a
    different server, for example using Invoke-Command,
    that sub-script may fail.

    If it does fail, the execute_ps method does not set
    the failed flag, so if the error stream isn't empty
    we have to assume failure and inspect all the streams.
    """
    logger.debug("there are error streams")
    errors = []
    for stream_collection in (stream for stream in (streams.error, streams.warning, streams.debug)):
        for s in stream_collection:
            if str(s) != s.message:
                errors.append(f"{str(s)}: {s.message}")
            else:
                errors.append(str(s))

    errors = list(set(errors))
    logger.debug(f"collected errors: {errors}")
    return errors


def _run_powershell(
    connection: Client,
    *,
    script: str,
    configuration_name: str = "Microsoft.PowerShell",
    environment: Optional[dict] = None,
) -> str:
    logger.info(f"executing powershell script with configuration_name {configuration_name}")
    std_out, streams, _ = connection.execute_ps(script, configuration_name=configuration_name, environment=environment)
    if streams.error:
        errors = _collect_errors(streams)
        if errors:
            raise exceptions.PowershellError(", ".join(errors))
        else:
            logger.debug(f"no error messages were found in available streams {str(streams)}")
            logger.debug(f"stdout: {std_out}")
            raise exceptions.PowershellError(
                "Script failed but no errors were found in the "
                "errors, warning, or debug streams. Please view "
                "the logs for more information."
            )
    else:
        return std_out


def _upload_file(
    connection: Client, *, source: str, destination: str, configuration_name: str = "Microsoft.PowerShell"
) -> str:
    try:
        return connection.copy(source, destination, configuration_name=configuration_name)
    except WinRMError as e:
        logger.error(e)
        winrm_error_string = "".join(e.args).replace("\n", "").replace("\r", "")
        if "DirectoryNotFoundException" in winrm_error_string:
            raise exceptions.PathError(f"Supplied destination path is invalid: '{destination}'")
        elif "being used by another process" in winrm_error_string:
            raise exceptions.CopyFileError(
                "Failed to upload file. The file is being used " "by another process on the server."
            )
        raise exceptions.CopyFileError("Failed to copy file to server. Please check the logs for more information.")
    except FileNotFoundError as e:
        logger.error(str(e))
        raise exceptions.PathError(f"Supplied source path does not exist: '{source}'")


def _download_file(
    connection: Client, *, source: str, destination: str, configuration_name: str = "Microsoft.PowerShell"
):
    try:
        return connection.fetch(source, destination, configuration_name=configuration_name)
    except WinRMError as e:
        logger.error(e)
        winrm_error_string = "".join(e.args).replace("\n", "").replace("\r", "")
        if re.search(re.compile(r"The\spath\sat\s.*\sdoes\snot\sexist"), repr(winrm_error_string)):
            raise exceptions.PathError(f"Supplied source path is invalid: '{source}'")
        raise exceptions.CopyFileError("Failed to copy file to server. Please check the logs for more information.")
    except FileNotFoundError as e:
        logger.error(str(e))
        raise exceptions.PathError(f"Supplied destination path does not exist: '{destination}'")
