import logging
import functools
import json

from collections import namedtuple
from typing import List, NamedTuple

from . import client

logger = logging.getLogger(__name__)

_header = ["Username", "Device", "Name", "State", "ID", "Type"]
_fields = ["username", "device", "sessionname", "state", "id", "type"]
_rdp_session = namedtuple("rdp_session", _fields)  # type: ignore

get_rdp_sessions = r"qwinsta | ForEach-Object {$_ -replace '\s{2,18}',',' } | ConvertFrom-CSV | ConvertTo-JSON"


def _as_objects(func):
    @functools.wraps(func)
    def _fn(connection, *args, **kwargs):
        output = func(connection, *args, **kwargs)
        if not output:
            return []
        json_data = json.loads(output)
        if not isinstance(json_data, list):
            json_data = [json_data]
        return [
            _rdp_session(  # type: ignore
                *[
                    session.get(field, "") for field in list(map(str.upper, _fields))
                ]  # qwinsta uses uppercase column names
            )
            for session in json_data
        ]

    return _fn


@_as_objects
def get_all_sessions(connection) -> List[NamedTuple]:
    """ Get all RDP sessions.

    :returns: A list of :ref:`rdp:RDP Records`.

    Example:

    .. code-block:: python

        with session() as connection:
            all_sessions = rdp.get_all_sessions(connection)

    """
    return client.run_powershell(connection, script=get_rdp_sessions)  # type: ignore


def get_active_sessions(connection) -> List[NamedTuple]:
    """ Get active RDP sessions.

    :returns: A list of :ref:`rdp:RDP Records`.

    Example:

    .. code-block:: python

        with session() as connection:
            active_sessions = rdp.get_all_sessions(connection)

    """
    return [s for s in get_all_sessions(connection) if getattr(s, "state").lower() == "active"]


def get_disconnected_sessions(connection) -> List[NamedTuple]:
    """ Get disconnected RDP sessions.

    :returns: A list of :ref:`rdp:RDP Records`.

    Example:

    .. code-block:: python

        with session() as connection:
            disconnected_sessions = rdp.get_disconnected_sessions(connection)

    """
    return [s for s in get_all_sessions(connection) if getattr(s, "state").lower() == "disc"]


def reset_session_by_id(connection, *, session_id) -> None:
    """ Reset an RDP session using its id.

    :param session_id: The id of the session to reset. When using :ref:`rdp:RDP Records`, the
        id can be accessed using dot notation.

    Example:

    .. code-block:: python

        with session() as connection:
            rdp.reset_session_by_id(connection, session.id)

    """
    return client.run_powershell(connection, script=f"rwinsta {session_id}")  # type: ignore
