"""
Group

This module contains actions for viewing & managing _local_ groups on a Windows Server.
"""
import logging

from typing import List, Dict, Any, Union

from ppa import ui
from windows_server import client

from .records import WindowsServer, User, UserData, Group, GroupData
from ._wrappers import winrm, is_domain_controller, json_load


logger = logging.getLogger(__name__)

GROUPS_TABLE_HEADER = ["Name", "Description"]
GROUPS_TABLE_FIELDS = [
    lambda u: u["name"],
    lambda u: u["description"] or "",
]


def _parse_groups(json_output: List[Dict[str, Any]]) -> List[Group]:
    return [GroupData(**group).__dict__ for group in json_output]


def _parse_users(json_output: List[Dict[str, Any]]) -> List[User]:
    return [UserData(**user).__dict__ for user in json_output]


@winrm
def get_all(windows_server: WindowsServer) -> Union[List[Group], list]:
    """Get all local groups from a Windows Server.

!!! question "Minimum Plugin Version: 2.0.0"

!!! danger "Domain Controllers"

    Running this action against a Domain Controller will return an empty list.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

??? success "Output"

    A list of [Group](../inputs_outputs/#group) dictionaries.

??? example

    ```YAML linenums="1"
    - windows_server.groups.get_all:
      load:
        windows_server: windows_server
      save: all_groups
    ```
"""
    if is_domain_controller(windows_server):
        return []

    output = client.run_powershell(
        windows_server, script="Get-WMIObject Win32_Group -Filter \"Domain='$env:COMPUTERNAME'\" | ConvertTo-JSON"
    )

    if not output:
        message = "Failed to get groups as the response from the server was empty."
        ui.output_error(message)
        raise ValueError(f"{message} Output: {repr(output)}.")

    return _parse_groups(json_load(output))


@winrm
def get_users(windows_server: WindowsServer, group_name: str) -> Union[List[User], list]:
    """Get all local users from a local group on a Windows Server.

!!! question "Minimum Plugin Version: 2.0.0"

!!! danger "Domain Controllers"

    Running this action against a Domain Controller will return an empty list.

!!! tip "Performance"

    This action uses WMI to get group members in an effort to avoid [this bug](https://github.com/PowerShell/PowerShell/issues/2996).

    It may take a few seconds on servers with many users & groups, or servers joined to a domain.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __group_name:__ the name of the group

??? success "Output"

    A list of [User](../inputs_outputs/#user) dictionaries.

??? example

    ```YAML linenums="1"
    - windows_server.groups.get_users:
        group_name: Remote Desktop Users
      load:
        windows_server: windows_server
      save: remote_desktop_users
    ```
"""
    if is_domain_controller(windows_server):
        return []

    output = client.run_powershell(
        windows_server,
        script=f"""(Get-WMIObject Win32_Group -Filter "Name=\'{group_name}\'").GetRelated("Win32_UserAccount") | `
Where-Object -Property "Domain" -EQ -Value $env:COMPUTERNAME | `
select Name, FullName, Description, Disabled, Domain, Lockout, PasswordChangeable, PasswordExpires, PasswordRequired, SID | `
ConvertTo-JSON""",
    )

    if not output:
        message = "Failed to get group members as the response from the server was empty."
        ui.output_error(message)
        raise ValueError(f"{message} Output: {repr(output)}.")

    return _parse_users(json_load(output))


def output_table(text: str, groups: Union[Group, List[Group]]) -> None:
    """Display groups in a table.

The table will have the following columns:

- Name
- Description

!!! question "Minimum Plugin Version: 1.7.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ a list of [Group](../inputs_outputs/#group) dictionaries to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.groups.get_all:
      load:
        windows_server: windows_server
      save: all_groups

    - windows_server.groups.output_table:
        text: All Local Groups
      load:
        groups: all_groups
    ```
    """
    ui.output_table(
        text, header=GROUPS_TABLE_HEADER, data=[[get(group) for get in GROUPS_TABLE_FIELDS] for group in groups],
    )


def output_tabbed_table(text: str, tabs: Dict[str, List[Group]]) -> None:
    """Display a groups table with multiple tabs.

This action can be used to display group lists from multiple servers.

The table will have the following columns:

- Name
- Description

!!! question "Minimum Plugin Version: 1.7.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is a list of [Group](../inputs_outputs/#group) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - set:
        name: tabbed_table
        value: {}

    - windows_server.groups.get_all:
      load:
        windows_server: windows_server
      save: all_groups

    - ppa_tools.dictionaries.insert:
        name: My Server
      load:
        value: all_groups
        dictionary: tabbed_table
      save: tabbed_table

    - windows_server.groups.output_tabbed_table:
        text: All Local Groups
      load:
        tabs: tabbed_table
    ```
    """
    ui.output_table(
        text,
        header=GROUPS_TABLE_HEADER,
        data={key: [[get(value) for get in GROUPS_TABLE_FIELDS] for value in value] for key, value in tabs.items()},
    )
