"""
Users

This module contains actions for auditing _local_ users on a Windows Server.
"""
import logging

from typing import List, Dict, Any, Union

from ppa import ui
from windows_server import client

from .records import WindowsServer, User, UserData
from ._wrappers import winrm, is_domain_controller, json_load


logger = logging.getLogger(__name__)

USERS_TABLE_HEADER = ["Name", "Full Name", "Description", "Disabled", "Locked"]
USERS_TABLE_FIELDS = [
    lambda u: u["name"],
    lambda u: u["full_name"] or "",
    lambda u: u["description"] or "",
    lambda u: str(u["disabled"]),
    lambda u: str(u["locked"]),
]


def _parse_users(json_output: List[Dict[str, Any]]) -> List[User]:
    return [UserData(**user).__dict__ for user in json_output]


@winrm
def get_all(windows_server: WindowsServer) -> Union[List[User], list]:
    """Get all local users from a Windows Server.

!!! question "Minimum Plugin Version: 2.0.0"

!!! danger "Domain Controllers"

    Running this action against a Domain Controller will return an empty list.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

??? success "Output"

    A list of [User](../inputs_outputs/#user) dictionaries.

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users
    ```
"""
    if is_domain_controller(windows_server):
        return []

    output = client.run_powershell(
        windows_server, script="Get-WmiObject -Class Win32_UserAccount -Filter \"LocalAccount='True'\" | ConvertTo-JSON"
    )

    if not output:
        message = "Failed to get users as the response from the server was empty."
        ui.output_error(message)
        raise ValueError(f"{message} Output: {repr(output)}.")

    return _parse_users(json_load(output))


def output_table(text: str, users: Union[User, List[User]]) -> None:
    """Display a list of users in a table.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 1.4.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ a list of [User](../inputs_outputs/#user) dictionaries to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - windows_server.users.output_table:
        text: All Local Users
      load:
        users: all_users
    ```
    """
    ui.output_table(
        text, header=USERS_TABLE_HEADER, data=[[get(user) for get in USERS_TABLE_FIELDS] for user in users],
    )


def output_tabbed_table(text: str, tabs: Dict[str, List[User]]) -> None:
    """Display a users table with multiple tabs.

This action can be used to display user lists from multiple servers.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 1.5.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is a list of [User](../inputs_outputs/#user) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - set:
        name: tabbed_table
        value: {}

    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - ppa_tools.dictionaries.insert:
        name: My Server
      load:
        value: all_users
        dictionary: tabbed_table
      save: tabbed_table

    - windows_server.users.output_tabbed_table:
        text: All Local Users
      load:
        tabs: tabbed_table
    ```
    """
    ui.output_table(
        text,
        header=USERS_TABLE_HEADER,
        data={key: [[get(value) for get in USERS_TABLE_FIELDS] for value in value] for key, value in tabs.items()},
    )
